# Copyright (c) Microsoft Corporation.  All rights reserved.

from .._sample_builder import SampleBuilderImpl, SampleResultImpl
from .._sampling.profiler import Profiler, ProfileTypes
from .._sampling.profile import DataMode
from collections import Counter


def profile(column):
    """A function for profiling a given column to identify the kind of data in the column,
    its distribution, and outliers.

    :param column: The column to learn a profile on.
    :type column: `Iterable[Any]`
    """
    profile_impl = Profiler(column_type=None).learn_profile(column)
    return DataProfile(profile_impl, column)


def sample(df):
    """A function for learning a sample from a given dataframe. The sample is intended to cover 
    both diversity in the common data values, as well as outliers.
    
    :param df: The dataframe to learn a sample on.
    :type df: `pandas.core.frame.DataFrame`
    """
    b = SampleBuilderImpl(df)
    return DataDiversitySample(b.learn(previous_result=None))


class DataDiversitySample(object):
    """The result from learning a sample for a given dataframe.

    (The result of calling :meth:`prose.datainsights.sample`.)"""

    def __init__(self, impl):
        self._impl = impl

    @property
    def dataframe(self):
        """A pandas dataframe containing a sample of the rows of the dataframe passed to 
        the builder.

        :rtype: `pandas.core.frame.DataFrame`
        """
        return self._impl.sample

    @property
    def preview(self):
        """A styled pandas dataframe containing a sample of the rows of the dataframe 
        passed to the builder, with the outliers highlighted.

        :rtype: `pandas.io.formats.style.Styler`
        """
        return self._impl.styled

    @property
    def outliers(self):
        """A dictionary mapping column names to the outlier values in the columns. The
        outliers include: (a) extreme values for numerical columns, (b) unusually frequent
        values for numerical columns indicative of missing values replaced with a default value,
        and (c) uncommon formats in mostly well-formatted semi-structured string columns.

        :rtype: `Dict[str, List[Any]]`
        """
        return self._impl.outliers

    def __repr__(self):
        """Return a printable string that summarizes the results of the sampling operation."""
        return repr(self._impl)


class DataProfile(object):
    def __init__(self, impl, data):
        self._impl = impl
        self._data = data
        self._outliers = None

    def covers(self, value):
        """Check if the current profile covers the given value."""
        return self._impl.covers(value)

    def signature(self, value):
        """Signature of the given value as per the current profile.

        :rtype `Tuple[bool, int]`
        """
        return self._impl.signature(value)

    @property
    def profile_type(self):
        """The type of profile.

        :rtype `ProfileTypes`
        """
        return self._impl.profile_type

    @property
    def outliers(self):
        """Set of values in the data deemed to be outliers.

        :rtype `Set[Any]`
        """
        if self._outliers is None:
            signatures = [self.signature(v) for v in self._data]
            freq = Counter(signatures)
            self._outliers = {
                v
                for (vi, v) in enumerate(self._data)
                if (not signatures[vi][0]) or (freq[signatures[vi]] == 1)
            }
        return self._outliers

    @property
    def modes(self):
        """List of different data modes in the profiles.
        """
        return self._impl.modes

    def __repr__(self):
        return repr(self._impl)
