# Copyright (c) Microsoft Corporation.  All rights reserved.

from collections import Counter
from ._sampling.profiler import Profiler
import pandas as pd
import math
import regex


class SampleBuilderImpl(object):
    """A builder to smartly sample from a given data frame.

    :param df: Data-frame to sample from.
    :type df: `pandas.core.frame.DataFrame`
    """

    _ExhaustivePerColumn = "ExhaustivePerColumn"
    _ExhaustiveAcrossColumns = "ExhaustiveAcrossColumns"

    @classmethod
    def _allowed_modes(cls):
        """Returns set of values allowed for `SampleBuilderImpl.mode`.

        :rtype: List[str]
        """
        return [
            SampleBuilderImpl._ExhaustivePerColumn,
            SampleBuilderImpl._ExhaustiveAcrossColumns,
        ]

    def __init__(self, df):
        self._df = df
        self._mode = SampleBuilderImpl._ExhaustivePerColumn
        self._target = "pandas"
        try:
            import pyspark

            if isinstance(self._df, pyspark.sql.DataFrame):
                self._target = "pyspark"
        except ImportError:
            pass

    @property
    def mode(self):
        """The mode in which the sampling is being conducted. This decides whether column
        correlations are considered during sampling, or not."""
        return self._mode

    @mode.setter
    def mode(self, value):
        """The mode in which the sampling is being conducted. This decides whether column
        correlations are considered during sampling, or not.

        :param value: The value to set for mode. Must be from SampleBuilderImpl._allowed_modes()
        :type value: str
        """
        allowed_modes = self._allowed_modes()
        if value not in allowed_modes:
            raise Exception(
                "Value not allowed for mode. Allowed modes: " + str(allowed_modes)
            )
        self._mode = value

    def _select(self, df, profilers, groups):
        """Select a sample dataframe from the given one, assuming the given profiles for 
        the columns, accounting for intra-group correlations.
        
        :param df: The data-frame to sample from.
        :type df: `pandas.core.frame.DataFrame`

        :param profilers: A dictionary mapping each column in the data-frame to a profiler for that column.
        :type profilers: Dict[str, Profiler]

        :param groups: A grouping of the data-frame columns. The correlations with each group is explored by the sample.
        :type groups: Iterable[Iterable[str]]

        :returns: A sampled data frame along with a styled object.
        :rtype: Tuple[pandas.core.frame.DataFrame, `pandas.io.formats.style.Styler`]
        """
        # Row signatures
        def row_signature(row):
            return [
                tuple([profilers[col].signature(row[col]) for col in group])
                for group in groups
            ]

        row_signatures = {i: row_signature(df.iloc[i]) for i in range(len(df))}

        # Memory full row scoring
        seen_signatures = [set() for i in range(len(groups))]

        def row_score(ri):
            score = 0
            row_sig = row_signatures[ri]
            for gi in range(len(groups)):
                if row_sig[gi] not in seen_signatures[gi]:
                    score += 1
            return score

        def select_row(ri):
            selected_sig = row_signatures[best]
            for gi in range(len(groups)):
                seen_signatures[gi].add(selected_sig[gi])

        # Sampling
        selected_rows = []
        while True:
            best, best_score = -1, 0
            for i in range(len(df)):
                score = row_score(i)
                if score > best_score:
                    best, best_score = i, score

            if best_score <= 0:
                break

            selected_rows.append(best)
            select_row(best)
        sample_df = df.iloc[selected_rows]

        group_frequencies = [
            Counter([rs[gi] for rs in row_signatures.values()])
            for gi in range(len(groups))
        ]
        column_to_group = {
            col: (gi, group.index(col))
            for col in df.columns
            for (gi, group) in enumerate(groups)
            if col in group
        }

        # Outliers
        def is_outlier(ri, gi, gii):
            freq = group_frequencies[gi][row_signatures[ri][gi]]
            return (not row_signatures[ri][gi][gii][0]) or (freq == 1)

        outliers = {col: set() for col in df.columns}
        for col in df.columns:
            gi, gii = column_to_group[col]
            outliers[col] = {
                df[col][ri] for ri in range(len(df)) if is_outlier(ri, gi, gii)
            }

        # Highlighting
        def highlight_outliers(column):
            gi, gii = column_to_group[column.name]

            return [
                "border: red; font-weight: bold; background-color: pink"
                if is_outlier(ri, gi, gii)
                else ""
                for ri in selected_rows
            ]

        styled = sample_df.style.apply(highlight_outliers).hide_index()
        return SampleResultImpl(profilers, sample_df, styled, outliers)

    def learn(self, previous_result=None):
        """Learn a sample for the data-frame the builder was constructed with.

        :param previous_result: If provided, the profilers from the previous result will be reused.
        :type previous_result: SampleResultImpl

        :returns: The results of sampling.
        :rtype: SampleResultImpl
        """
        df = self._df

        # Construct/get profilers
        if previous_result is not None:
            profilers = previous_result._profilers
        else:
            # TODO: Do not rely on pandas data-type detection. It is flaky when it comes
            # to integer columns with nulls in them.
            profilers = {col: Profiler(df.dtypes[col]) for col in df.columns}

        # Profile all columns
        for col in df.columns:
            profiler = profilers[col]
            if self._target == "pyspark":
                profiler.learn_profile_pyspark(df, df[col])
            else:
                profiler.learn_profile(df[col])

        # Sample with sampling mode
        if self.mode == SampleBuilderImpl._ExhaustivePerColumn:
            return self._select(df, profilers, [[col] for col in df.columns])
        elif self.mode == SampleBuilderImpl._ExhaustiveAcrossColumns:
            return self._select(df, profilers, [list(df.columns)])
        else:
            # This should be unreachable currently.
            # TODO: In the future, add a mode to automatically find correlations and group columns based on that.
            raise NotImplementedError

    def __repr__(self):
        return """SampleBuilder
==================
input dataframe: {0} rows
columns: {1}
mode: {2}""".format(
            len(self._df), list(self._df.columns), self.mode
        )


class SampleResultImpl(object):
    """Represents the results of a smart sample operation. """

    def __init__(self, profilers, sample, styled, outliers):
        self._profilers = profilers
        self.sample = sample
        self.styled = styled
        self.outliers = outliers

    def profiler_for(self, column):
        """The profiler which was used for the given column when constructing the sample.

        :param column: The column for which to return the profiler.
        :type column: str

        :rtype: Profiler
        """
        return self._profilers[column]

    def __repr__(self):
        return """SampleResult
==================
sampled dataframe: {0} rows""".format(
            len(self.sample)
        )
