# Copyright (c) Microsoft Corporation.  All rights reserved.

from .profile import Profile, ProfileTypes, ProfilingStrategy, DataMode
import math


class Category(DataMode):
    def __init__(self, value):
        super().__init__("Category[%s]" % str(value), [value])
        self.value = value


class CategoricalProfile(Profile):
    """A profile that classifies a column as categorical data.

    :param categories: The categories in the given data.
    :type categories: List[object]

    :param active: Is the profile currently valid?
    :type active: bool
    """

    _MAX_CATEGORIES = 100
    _CATEGORICAL_LIMIT = 0.1

    def __init__(self, categories, active=True):
        super().__init__(ProfileTypes.CATEGORICAL, ProfilingStrategy.ITERATIVE)
        self._active = active
        self.categories = categories
        self.has_nan_category = any(not (c == c) for c in categories)
        self.supports_pyspark = True

    def covers(self, value):
        """Check if the given value is covered under the known categories.

        :param value: The value to be checked.
        :returns: A boolean if the value is a known categorical value.
        :rtype: bool
        """
        if (not (value == value)) and self.has_nan_category:
            return True
        return value in self.categories

    @property
    def active(self):
        """Check if the profile is active, i.e., valid.

        :returns: A boolean if the profile is currently valid.
        :rtype: bool
        """
        return self._active

    @property
    def modes(self):
        return [Category(c) for c in self.categories]

    def signature(self, value):
        """Return a signature of the given value as per this profile. The signature
        currently involves whether a value is covered, and what category it falls into.

        :returns: A signature of the given value.
        :rtype: Tuple[bool, int]
        """
        for (ci, cat) in enumerate(self.categories):
            if cat == value:
                return (True, ci)
        return (False, -1)

    def with_values(self, new_values, processed):
        """Produce a new categorical profile along with the newly given values.

        :returns: A profile with categories from the newly given values.
        :rtype: CategoricalProfile
        """
        has_nan = False
        new_values_filtered = list(v for v in new_values if v == v)
        if len(new_values_filtered) != len(new_values):
            has_nan = True

        categories = list(set(new_values_filtered) | set(self.categories))
        if not self.has_nan_category and has_nan:
            categories += [math.nan]
        if len(categories) > CategoricalProfile._MAX_CATEGORIES:
            return None
        if len(categories) <= CategoricalProfile._CATEGORICAL_LIMIT * processed:
            return CategoricalProfile(categories, active=True)
        else:
            return CategoricalProfile(categories, active=False)

    def _with_values_pyspark(self, df, column):
        # We need both dataframe and column for pyspark.
        # There are three reasonable ways of doing this:
        # 1. Use approxDistinctCount to check the number, and then use a
        #    distinct on the column to get the actual values.
        # 2. Use distinct and collect with a limit.
        # 3. Use cegis like iteration.
        # Here, we are using #3. This is more likely to fail fast on columns
        # that are not categorical.

        remaining_values = df.select(column)
        categories = []
        while True:
            new_categories = [
                row[0]
                for row in remaining_values.sample(False, 0.1, 0)
                .limit(CategoricalProfile._MAX_CATEGORIES)
                .collect()
            ]
            if len(new_categories) == 0:
                break
            categories = list(set(new_categories) | set(categories))
            if len(categories) > CategoricalProfile._MAX_CATEGORIES:
                return None
            remaining_values = remaining_values.where(
                column.isin(new_categories) == False
            )
        size = df.count()
        if len(categories) <= CategoricalProfile._CATEGORICAL_LIMIT * size:
            return CategoricalProfile(categories, active=True)
        else:
            return CategoricalProfile(categories, active=False)

    def __repr__(self):
        """A string representation of the profile

        :rtype: str
        """
        return (
            "Categorical["
            + ", ".join([str(c) if c is not None else "None" for c in self.categories])
            + "]"
        )
