# Copyright (c) Microsoft Corporation.  All rights reserved.

import statistics
from scipy import stats, signal
from .profile import Profile, ProfileTypes, ProfilingStrategy, DataMode
from collections import Counter
from enum import Enum
import numpy as np
import math


class NumericMode(Enum):
    NORMAL = 1
    SINGLETON = 2
    MINMAX = 3


class Mode(DataMode):
    """An object that represents a single mode within a distribution.

    :param values: Values from which the mode was constructed.
    :type values: Iterable[int]

    :param kind: The kind of mode.
    :type kind: NumericMode
    """

    def __init__(self, values, kind):
        self.values = values
        self.min = min(values)
        self.max = max(values)
        super().__init__(
            "Numeric[%s, %s]" % (str(self.min), str(self.max)),
            [self.min] + values[:3] + [self.max],
        )
        self.mean = statistics.mean(values)
        if len(values) >= 2:
            self.stdev = statistics.stdev(values)
        else:
            self.stdev = 0
        self.kind = kind

    @classmethod
    def from_values(cls, values, kind=NumericMode.NORMAL):
        """Construct a Mode from a list of given values.

        :param values: Values from which to construct a mode.
        :type values: Iterable[int]

        :param kind: The kind of mode.
        :type kind: NumericMode

        :rtype: Mode
        """
        return Mode(values, kind)

    @property
    def allowed_range(self):
        """Return the minimum and maximum values that are covered by this mode.

        :rtype: Tuple[int, int]
        """
        if self.kind == NumericMode.NORMAL:
            return ((self.mean - 5 * self.stdev), (self.mean + 5 * self.stdev))
        elif self.kind == NumericMode.SINGLETON:
            return (self.min, self.max)
        elif self.kind == NumericMode.MINMAX:
            return (self.min, self.max)
        else:
            raise NotImplementedError("Unknown numeric mode: %s" % str(self.kind))

    def __contains__(self, value):
        """Return if the given value is covered by this mode.

        :param value: The value to check.
        :type value: int

        :rtype: bool
        """
        if self.kind == NumericMode.SINGLETON and math.isnan(self.min):
            return math.isnan(value)
        (lower, upper) = self.allowed_range
        return lower <= value <= upper

    def __repr__(self):
        """A string representation of the profile

        :rtype: str
        """
        (lower, upper) = self.allowed_range
        return "[%s, %s]" % (str(lower), str(upper))


class NumericProfile(Profile):
    """A profile that classifies a column as numeric data with a few known numerical "modes".

    :param modes: The numeric modes present in the data.
    :type modes: Iterable[Mode]

    :param active: Is the current profile valid?
    :type active: bool
    """

    def __init__(self, modes, active=True):
        super().__init__(ProfileTypes.NUMERIC, ProfilingStrategy.FULL)
        self._modes = modes
        self._active = active

    def covers(self, value):
        """Check if the given value falls into one of the known modes.

        :returns: True if the value is covered under a known mode.
        :rtype: bool
        """
        return any(value in mode for mode in self.modes)

    @property
    def modes(self):
        return self._modes

    @property
    def active(self):
        """Is the profile currently valid? 

        :returns: True if the profile is active.
        :rtype: bool
        """
        return self._active

    def signature(self, value):
        """Return a signature of the given value as per this profile. The signature
        currently involves whether a value is covered, and what mode it is covered by.

        :returns: A signature of the given value.
        :rtype: Tuple[bool, int]
        """
        for (mi, mode) in enumerate(self.modes):
            if value in mode:
                return (mode.min != mode.max, mi)
        return (False, -1)

    def _frequent_value_analysis(self, values):
        if len(values) < 2:
            return ([], values)

        freq_map = Counter(values)
        (lower, upper, mean, stdev) = (
            min(values),
            max(values),
            statistics.mean(values),
            statistics.stdev(values),
        )

        def get_granularity(v):
            vd = list(sorted(set(v)))
            first_quartile = vd[(len(vd) * 1) // 4]
            third_quartile = vd[(len(vd) * 3) // 4]
            if (len(vd) // 2) != 0:
                return (third_quartile - first_quartile) / (len(vd) // 2)
            return None

        frequent_modes = []

        range_length = upper - lower
        if upper == lower:
            mode = Mode.from_values(values, NumericMode.SINGLETON)
            return [mode], []

        granularity = get_granularity(values)
        if granularity is None:
            return [], values

        expected_distinct_values = 1 + (
            range_length / granularity
        )  # Number of distinct values expected
        expected_mean_count = len(values) / expected_distinct_values

        for (value, count) in freq_map.most_common(5):
            # We use a staggered limiting strategy.
            # If the value is closer to the mean, the more common it should be to be suspicious.
            # The limits (100, 50, and 30) were arrived at empirically by testing on several datasets.

            # Between mu + sigma, 100 times more common than usual
            if (mean - stdev <= value <= mean + stdev) and (
                count >= expected_mean_count * 100
            ):
                # print(value, "< 1 sigma from mean")
                frequent_modes.append(
                    Mode.from_values([value] * count, NumericMode.SINGLETON)
                )
                values = [v for v in values if v != value]
            elif (mean - 2 * stdev <= value <= mean + 2 * stdev) and (
                count >= expected_mean_count * 50
            ):
                # print(value, "< 2 sigma from mean")
                frequent_modes.append(
                    Mode.from_values([value] * count, NumericMode.SINGLETON)
                )
                values = [v for v in values if v != value]
            elif count >= expected_mean_count * 30:
                # print(value, "> 2 sigma from mean")
                frequent_modes.append(
                    Mode.from_values([value] * count, NumericMode.SINGLETON)
                )
                values = [v for v in values if v != value]

        return (frequent_modes, values)

    def _modality_analysis(self, values):
        (lower, upper) = min(values), max(values)
        length = upper - lower
        if length == 0:
            return [Mode.from_values(values)]

        # Perform kernel density estimation
        kde = stats.gaussian_kde(values)
        domain = np.linspace(lower - length / 10, upper + length / 10, num=500)
        y = kde(domain)

        # Local minima and maxima in the kde
        peaks = list(signal.argrelextrema(y, np.greater_equal)[0])
        troughs = list(signal.argrelextrema(y, np.less_equal)[0])

        # argrelextrema with np.less_equal returns the end points of the domain
        # if the function is increasing/decreasing at the left/right end point.
        # We eliminate those and later use negative/positive infinity.
        if len(troughs) > 0 and troughs[0] == 0:
            troughs = troughs[1:]
        if len(troughs) > 0 and troughs[-1] == len(domain) - 1:
            troughs = troughs[:-1]

        # Modes are between two troughs (with negative and positive infinity)
        # Here, the check m1 != m0 + 1 is to eliminate plateau minima in the KDE.
        modes = [
            (m0, m1)
            for (m0, m1) in zip([-math.inf] + troughs, troughs + [math.inf])
            if m1 != m0 + 1
        ]

        # Generally, modes are bounded by minima in the KDE
        # However, if the peaks in the modes are not prominent, we merge neighbouring modes
        mode_peaks = {
            mode: next(p for p in peaks if mode[0] <= p <= mode[1]) for mode in modes
        }
        mi = 0
        while mi < len(modes) - 1:
            mode1, mode2 = modes[mi], modes[mi + 1]
            p1 = y[mode_peaks[mode1]]
            p2 = y[mode_peaks[mode2]]
            t = y[mode1[1]]
            if t > 0.8 * min(p1, p2):
                merged_mode = (mode1[0], mode2[1])
                modes[mi : mi + 2] = [merged_mode]
                mode_peaks[merged_mode] = (
                    mode_peaks[mode1] if p1 > p2 else mode_peaks[mode2]
                )
                mi -= 1
            mi += 1

        if len(modes) > 4:
            return None

        # Translate to our mode data-structure
        rmodes = []
        for mode in modes:
            l = domain[mode[0]] if mode[0] > -math.inf else -math.inf
            u = domain[mode[1]] if mode[1] < math.inf else math.inf
            mode_values = [v for v in values if l <= v < u]
            if len(mode_values) > 0.05 * len(values):
                rmodes.append(Mode.from_values(mode_values))
        return rmodes

    # For now, something very stupid.
    def with_values(self, new_values, processed):
        """Produce a new NumericProfile with modes for the new values.

        :returns: A pattern profile that covers the additionally provided values.
        :rtype: NumericProfile
        """
        if any(map(lambda v: v is not None and not np.isreal(v), new_values)):
            return NumericProfile([], active=False)

        values = list(new_values) + [v for mode in self.modes for v in mode.values]
        modes = []

        # Look for nans
        l = len(values)
        values = [v for v in values if not math.isnan(v)]
        nl = len(values)
        if l != nl:
            modes.append(
                Mode.from_values([float("nan")] * (l - nl), NumericMode.SINGLETON)
            )

        # Look for frequent values (these are likely to be errors).
        frequent_modes, values = self._frequent_value_analysis(values)
        modes.extend(frequent_modes)

        if len(values) > 0:
            new_modes = self._modality_analysis(values)
            if new_modes is not None:
                # Clean multi-modes
                modes.extend(new_modes)
            else:
                # Otherwise, assume single distribution.
                modes.append(Mode.from_values(values))
        return NumericProfile(modes, active=True)

    def __repr__(self):
        """A string representation of the profile

        :rtype: str
        """
        return "Numerical[ " + ", ".join([str(m) for m in self.modes]) + " ]"
