# Copyright (c) Microsoft Corporation.  All rights reserved.

from .profile import Profile, ProfileTypes, ProfilingStrategy, DataMode
from .. import _internals
from itertools import islice
import regex


class Pattern(DataMode):
    def __init__(self, regex, examples):
        super().__init__(description="Regex[%s]" % regex.pattern, examples=examples)
        self.regex = regex


class PatternProfile(Profile):
    """A profile that classifies a column as string data with a few known regex patterns.

    :param values: The values from which the patterns were learned.
    :type values: Iterable[str]

    :param regexes: The regex patterns that are present in this column.
    :type regexes: Iterable[regex.Regex]

    :param highlights: A boolean representing whether each pattern should be considered an outlier.
    :type highlights: Iterable[bool]

    :param active: A boolean indicating whether the profile is currently valid.
    :type active: bool
    """

    _cegis_iteration_limit = 20
    _long_string_limit = 40
    _incremental_sample_size = 100

    def __init__(self, values, regexes, highlights, active=True):
        super().__init__(ProfileTypes.PATTERNS, ProfilingStrategy.ITERATIVE)
        self.values = values
        self.regexes = regexes
        self.highlights = highlights
        self._active = active
        self.supports_pyspark = True
        self._modes = None

    @property
    def modes(self):
        def get_examples(r):
            return list(islice((v for v in self.values if r.fullmatch(v)), 3))

        if self._modes is None:
            self._modes = [
                Pattern(regex, get_examples(regex)) for regex in self.regexes
            ]
        return self._modes

    def covers(self, value):
        """Check if the given value is covered by one of the known regex patterns.

        :returns: True if the value is covered under a known pattern.
        :rtype: bool
        """
        try:
            for regex in self.regexes:
                if regex.fullmatch(value):
                    return True
            return False
        except:
            return False

    @property
    def active(self):
        """Is the profile currently valid? 

        :returns: True if the profile is active.
        :rtype: bool
        """
        return self._active

    def _with_values_pyspark(self, df, column):
        patterns = []
        highlights = []
        values = []
        for i in range(self._cegis_iteration_limit):
            remaining_values = df.select(column)
            for pattern in patterns:
                remaining_values = remaining_values.filter(
                    column.rlike(pattern) == False
                )
            new_values = [
                row[0]
                for row in remaining_values.limit(
                    self._incremental_sample_size
                ).collect()
            ]
            if len(new_values) == 0:
                break
            values += new_values

            all_data = [v for v in (self.values + values) if isinstance(v, str)]
            pattern_info = self._learn_patterns(all_data)
            if pattern_info is None:
                return None
            patterns, highlights = pattern_info
        return PatternProfile(all_data, [regex.Regex(r) for r in patterns], highlights)

    def _learn_patterns(self, data):
        if any(
            map(lambda s: isinstance(s, str) and len(s) > self._long_string_limit, data)
        ):
            return None
        try:
            f = _internals.PatternFinder(data)
            new_regexes = f.find()
            highlights = [False] * len(new_regexes)
        except:
            return None
        return new_regexes, highlights

    def with_values(self, new_values, processed):
        """Produce a new PatternProfile with patterns for the new values.

        :returns: A pattern profile that covers the additionally provided values.
        :rtype: PatternProfile
        """
        if any(map(lambda v: v is not None and type(v) != str, new_values)):
            return None

        all_data = [v for v in (self.values + new_values) if isinstance(v, str)]
        pattern_info = self._learn_patterns(all_data)
        if pattern_info is None:
            return None
        new_regexes, highlights = pattern_info
        return PatternProfile(
            all_data, [regex.Regex(r) for r in new_regexes], highlights
        )

    def signature(self, value):
        """Return a signature of the given value as per this profile. The signature
        currently involves whether a value is covered, and what regex pattern it falls
        into.

        :returns: A signature of the given value.
        :rtype: Tuple[bool, int]
        """
        try:
            for (ri, reg) in enumerate(self.regexes):
                if reg.fullmatch(value):
                    return (True, ri)
            return (False, -1)
        except:
            return (False, -2)

    def __repr__(self):
        """A string representation of the profile

        :rtype: str
        """
        return "Patterns[ " + ", ".join([r.pattern for r in self.regexes]) + " ]"
