# Copyright (c) Microsoft Corporation.  All rights reserved.

from enum import Enum
from abc import ABC, abstractmethod


class ProfileTypes(Enum):
    """An enumeration of types of profiles."""

    NUMERIC = 1
    PATTERNS = 2
    CATEGORICAL = 3
    NONE = 4


class ProfilingStrategy(Enum):
    """An enumeration of strategies for building profiles."""

    ITERATIVE = 1
    FULL = 2


class DataMode(ABC):
    """A single mode in a given column."""

    def __init__(self, description, examples):
        self.description = description
        self.examples = examples


class Profile(ABC):
    """An abstract class that represents all the data in a given column.
    
    :param profile_type: The type of the profile.
    :type profile_type: ProfileTypes

    :param profiling_strategy: The strategy to build this profile.
    :type profiling_strategy: ProfilingStrategy
    """

    def __init__(self, profile_type, profiling_strategy):
        self.profile_type = profile_type
        self.profiling_strategy = profiling_strategy
        self.supports_pyspark = False

    @abstractmethod
    def covers(self, value):
        """Check if the current profile covers the given value."""

    @abstractmethod
    def with_values(self, new_values, processed_count):
        """Generate a new profile of the same kind that also covers new_values."""

    @property
    @abstractmethod
    def active(self):
        """Is this a valid profile currently?"""

    @abstractmethod
    def signature(self, value):
        """Signature of the given value as per the current profile."""

    @property
    @abstractmethod
    def modes(self):
        """Modes in the learned profile."""

    def filter_uncovered(self, values, limit):
        """Produce the values that are not covered by the current profile, up to limit number of values. 

        :rtype: Iterable[object]
        """
        processed = 0
        count = 0
        uncovered = []
        for value in values:
            processed += 1
            if not self.covers(value):
                uncovered.append(value)
                count += 1
            if count >= limit:
                break
        return (uncovered, processed)
