# Copyright (c) Microsoft Corporation.  All rights reserved.

from .profile import ProfileTypes, ProfilingStrategy
from .numeric_profile import NumericProfile
from .pattern_profile import PatternProfile
from .default_profile import DefaultProfile
from .categorical_profile import CategoricalProfile


class Profiler(object):
    """An object is can learn a profile for a given numerical, string, or categorical column.
    
    :param column_type: A string representation of the data-type of the column.
    :type column_type: str
    """

    _batch_size = 100

    def __init__(self, column_type=None):
        if column_type is None:
            self._candidate_profiles = [
                CategoricalProfile([]),
                NumericProfile([]),
                PatternProfile([], [], []),
                DefaultProfile(),
            ]
        elif column_type == "int64" or column_type == "float64":
            self._candidate_profiles = [
                CategoricalProfile([]),
                NumericProfile([]),
                DefaultProfile(),
            ]
        else:
            self._candidate_profiles = [
                CategoricalProfile([]),
                PatternProfile([], [], []),
                DefaultProfile(),
            ]

    def learn_profile_pyspark(self, df, column):
        self._candidate_profiles = [
            profile for profile in self._candidate_profiles if profile.supports_pyspark
        ]

        new_profiles = []
        for profile in self._candidate_profiles:
            profile = profile._with_values_pyspark(df, column)
            if profile is not None:
                new_profiles.append(profile)
        self._candidate_profiles = new_profiles
        return self.profile

    def learn_profile(self, values):
        """Learn a profile for the given values from a column.

        :param values: The values from which to learn a profile.
        :type values: Iterable[object]
        """
        new_profiles = []
        for profile in self._candidate_profiles:
            if profile.profiling_strategy == ProfilingStrategy.FULL:
                profile = profile.with_values(values, processed=None)
                new_profiles.append(profile)
                if profile.active:
                    break
            elif profile.profiling_strategy == ProfilingStrategy.ITERATIVE:
                iterations = 0
                (uncovered, processed) = profile.filter_uncovered(
                    values, Profiler._batch_size
                )
                while iterations < 20 and len(uncovered) > 0:
                    profile = profile.with_values(uncovered, processed)
                    if profile is None:
                        break
                    (uncovered, processed) = profile.filter_uncovered(
                        values, Profiler._batch_size
                    )
                    iterations += 1
                if profile is not None:
                    new_profiles.append(profile)
                    if profile.active:
                        break
            else:
                raise NotImplementedError
        self._candidate_profiles = new_profiles
        return self.profile

    def covers(self, value):
        """Check if the given value covered by the currently active profile.
        
        :param value: Value to be checked.
        :type value: object

        :returns: True if the value is covered by the active profile.
        :rtype: bool
        """
        current_profile = self.profile
        return (current_profile is not None) and current_profile.covers(value)

    @property
    def profile(self):
        """The best profile that can be learned from the data that has been seen.

        :rtype: Profile
        """
        for profile in self._candidate_profiles:
            if profile.active:
                return profile
        return None

    def signature(self, value):
        """Signature of the value according to the currently active profile. Values with 
        the same signature are considered similar.

        :param value: The value on which to compute signature
        :type value: object

        :returns: The signature of the value.
        :rtype: Tuple[bool, object]
        """
        return self.profile.signature(value)
